"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    """

"""Functions to calculate nutation and obliquity values.

The IAU "1980 Theory of Nutation" is used, but terms with coefficients smaller
than 0.0003" have been dropped.

Reference: Jean Meeus, _Astronomical Algorithms_, second edition, 1998,
Willmann-Bell, Inc.

The first edition of the Meeus book had some errors in the table. These may be
corrected in the second edition. I recall correcting my values from those
published in _Explanatory Supplement to the Astronomical Almanac_, revised
edition edited by P. Kenneth Seidelman, 1992

"""

import numpy as np

from astronomia.util import polynomial, modpi2, d_to_r, dms_to_d
from astronomia.calendar import jd_to_jcent

from astronomia.commonterms import kD, kM, kM1, kF, ko

# [Meeus-1998: table 22.A]
#
#    D, M, M1, F, omega, psiK, psiT, epsK, epsT

_tbl = (
    ( 0,  0,  0,  0,  1, -171996, -1742, 92025,  89),
    (-2,  0,  0,  2,  2,  -13187,   -16,  5736, -31),
    ( 0,  0,  0,  2,  2,   -2274,    -2,   977,  -5),
    ( 0,  0,  0,  0,  2,    2062,     2,  -895,   5),
    ( 0,  1,  0,  0,  0,    1426,   -34,    54,  -1),
    ( 0,  0,  1,  0,  0,     712,     1,    -7,   0),
    (-2,  1,  0,  2,  2,    -517,    12,   224,  -6),
    ( 0,  0,  0,  2,  1,    -386,    -4,   200,   0),
    ( 0,  0,  1,  2,  2,    -301,     0,   129,  -1),
    (-2, -1,  0,  2,  2,     217,    -5,   -95,   3),
    (-2,  0,  1,  0,  0,    -158,     0,     0,   0),
    (-2,  0,  0,  2,  1,     129,     1,   -70,   0),
    ( 0,  0, -1,  2,  2,     123,     0,   -53,   0),
    ( 2,  0,  0,  0,  0,      63,     0,     0,   0),
    ( 0,  0,  1,  0,  1,      63,     1,   -33,   0),
    ( 2,  0, -1,  2,  2,     -59,     0,    26,   0),
    ( 0,  0, -1,  0,  1,     -58,    -1,    32,   0),
    ( 0,  0,  1,  2,  1,     -51,     0,    27,   0),
    (-2,  0,  2,  0,  0,      48,     0,     0,   0),
    ( 0,  0, -2,  2,  1,      46,     0,   -24,   0),
    ( 2,  0,  0,  2,  2,     -38,     0,    16,   0),
    ( 0,  0,  2,  2,  2,     -31,     0,    13,   0),
    ( 0,  0,  2,  0,  0,      29,     0,     0,   0),
    (-2,  0,  1,  2,  2,      29,     0,   -12,   0),
    ( 0,  0,  0,  2,  0,      26,     0,     0,   0),
    (-2,  0,  0,  2,  0,     -22,     0,     0,   0),
    ( 0,  0, -1,  2,  1,      21,     0,   -10,   0),
    ( 0,  2,  0,  0,  0,      17,    -1,     0,   0),
    ( 2,  0, -1,  0,  1,      16,     0,    -8,   0),
    (-2,  2,  0,  2,  2,     -16,     1,     7,   0),
    ( 0,  1,  0,  0,  1,     -15,     0,     9,   0),
    (-2,  0,  1,  0,  1,     -13,     0,     7,   0),
    ( 0, -1,  0,  0,  1,     -12,     0,     6,   0),
    ( 0,  0,  2, -2,  0,      11,     0,     0,   0),
    ( 2,  0, -1,  2,  1,     -10,     0,     5,   0),
    ( 2,  0,  1,  2,  2,      -8,     0,     3,   0),
    ( 0,  1,  0,  2,  2,       7,     0,    -3,   0),
    (-2,  1,  1,  0,  0,      -7,     0,     0,   0),
    ( 0, -1,  0,  2,  2,      -7,     0,     3,   0),
    ( 2,  0,  0,  2,  1,      -7,     0,     3,   0),
    ( 2,  0,  1,  0,  0,       6,     0,     0,   0),
    (-2,  0,  2,  2,  2,       6,     0,    -3,   0),
    (-2,  0,  1,  2,  1,       6,     0,    -3,   0),
    ( 2,  0, -2,  0,  1,      -6,     0,     3,   0),
    ( 2,  0,  0,  0,  1,      -6,     0,     3,   0),
    ( 0, -1,  1,  0,  0,       5,     0,     0,   0),
    (-2, -1,  0,  2,  1,      -5,     0,     3,   0),
    (-2,  0,  0,  0,  1,      -5,     0,     3,   0),
    ( 0,  0,  2,  2,  1,      -5,     0,     3,   0),
    (-2,  0,  2,  0,  1,       4,     0,     0,   0),
    (-2,  1,  0,  2,  1,       4,     0,     0,   0),
    ( 0,  0,  1, -2,  0,       4,     0,     0,   0),
    (-1,  0,  1,  0,  0,      -4,     0,     0,   0),
    (-2,  1,  0,  0,  0,      -4,     0,     0,   0),
    ( 1,  0,  0,  0,  0,      -4,     0,     0,   0),
    ( 0,  0,  1,  2,  0,       3,     0,     0,   0),
    ( 0,  0, -2,  2,  2,      -3,     0,     0,   0),
    (-1, -1,  1,  0,  0,      -3,     0,     0,   0),
    ( 0,  1,  1,  0,  0,      -3,     0,     0,   0),
    ( 0, -1,  1,  2,  2,      -3,     0,     0,   0),
    ( 2, -1, -1,  2,  2,      -3,     0,     0,   0),
    ( 0,  0,  3,  2,  2,      -3,     0,     0,   0),
    ( 2, -1,  0,  2,  2,      -3,     0,     0,   0))


def _constants(T):
    """Return some values needed for both nutation_in_longitude() and
    nutation_in_obliquity()"""
    D = modpi2(polynomial(kD,  T))
    M = modpi2(polynomial(kM,  T))
    M1 = modpi2(polynomial(kM1, T))
    F = modpi2(polynomial(kF,  T))
    omega = modpi2(polynomial(ko,  T))
    return D, M, M1, F, omega


def nutation_in_longitude(jd):
    """Return the nutation in longitude.

    High precision. [Meeus-1998: pg 144]

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - nutation in longitude, in radians

    """
    #
    # Future optimization: factor the /1e5 and /1e6 adjustments into the table.
    #
    # Could turn the loop into a generator expression. Too messy?
    #
    T = jd_to_jcent(jd)
    D, M, M1, F, omega = _constants(T)
    deltaPsi = 0.0
    for tD, tM, tM1, tF, tomega, tpsiK, tpsiT, tepsK, tepsT in _tbl:
        arg = D*tD + M*tM + M1*tM1 + F*tF + omega*tomega
        deltaPsi += (tpsiK/10000.0 + tpsiT/100000.0 * T) * np.sin(arg)

    deltaPsi /= 3600
    deltaPsi = d_to_r(deltaPsi)
    return deltaPsi


def nutation_in_obliquity(jd):
    """Return the nutation in obliquity.

    High precision. [Meeus-1998: pg 144]

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - nutation in obliquity, in radians

    """
    #
    # Future optimization: factor the /1e5 and /1e6 adjustments into the table.
    #
    # Could turn the loop into a generator expression. Too messy?
    #
    T = jd_to_jcent(jd)
    D, M, M1, F, omega = _constants(T)
    deltaEps = 0.0
    for tD, tM, tM1, tF, tomega, tpsiK, tpsiT, tepsK, tepsT in _tbl:
        arg = D*tD + M*tM + M1*tM1 + F*tF + omega*tomega
        deltaEps = deltaEps + (tepsK/10000.0 +
                               tepsT/100000.0 * T) * np.cos(arg)
    deltaEps = deltaEps / 3600
    deltaEps = d_to_r(deltaEps)
    return deltaEps

#
# Constant terms
#
_el0 = (d_to_r(dms_to_d(23, 26, 21.448)),
        d_to_r(dms_to_d(0, 0, -46.8150)),
        d_to_r(dms_to_d(0, 0, -0.00059)),
        d_to_r(dms_to_d(0, 0, 0.001813)))


def obliquity(jd):
    """Return the mean obliquity of the ecliptic.

    Low precision, but good enough for most uses. [Meeus-1998: equation 22.2].
    Accuracy is 1" over 2000 years and 10" over 4000 years.

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - obliquity, in radians

    """
    T = jd_to_jcent(jd)
    return polynomial(_el0, T)


#
# Constant terms
#
_el1 = (d_to_r(dms_to_d(23, 26, 21.448)),
        d_to_r(dms_to_d(0, 0, -4680.93)),
        d_to_r(dms_to_d(0, 0, -1.55)),
        d_to_r(dms_to_d(0, 0, 1999.25)),
        d_to_r(dms_to_d(0, 0, -51.38)),
        d_to_r(dms_to_d(0, 0, -249.67)),
        d_to_r(dms_to_d(0, 0, -39.05)),
        d_to_r(dms_to_d(0, 0, 7.12)),
        d_to_r(dms_to_d(0, 0, 27.87)),
        d_to_r(dms_to_d(0, 0, 5.79)),
        d_to_r(dms_to_d(0, 0, 2.45)))


def obliquity_hi(jd):
    """Return the mean obliquity of the ecliptic.

    High precision [Meeus-1998: equation 22.3].

    Accuracy is 0.01" between 1000 and 3000, and "a few arc-seconds
    after 10,000 years".

    Arguments:
      - `jd` : Julian Day in dynamical time

    Returns:
      - obliquity, in radians

    """
    U = jd_to_jcent(jd) / 100
    return polynomial(_el1, U)




from __future__ import division

"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    """

"""A collection of date and time functions.

The functions which use Julian Day Numbers are valid only for positive values,
i.e., for dates after -4712 (4713BC).

Unless otherwise specified, Julian Day Numbers may be fractional values.

Numeric years use the astronomical convention of a year 0: 0 = 1BC, -1 = 2BC,
etc.

Numeric months are 1-based: Jan = 1...Dec = 12.

Numeric days are the same as the calendar value.

Reference: Jean Meeus, _Astronomical Algorithms_, second edition, 1998,
Willmann-Bell, Inc.
"""

from math import modf

import numpy as np

from astronomia.util import d_to_r, modpi2, _scalar_if_one
from astronomia.constants import minutes_per_day, seconds_per_day
import astronomia.globals


class Error(Exception):
    """local exception class"""
    pass


def frac_yr_to_jd(year, gregorian=True):
    """Convert a date in the Julian or Gregorian fractional year to the
    Julian Day Number (Meeus 7.1).

    Arguments:
      - `year` : (int, float)  year

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (float)

    """
    year = np.atleast_1d(year)
    day = np.atleast_1d(0.0).astype(np.float64)
    year, day = map(np.array, np.broadcast_arrays(year, day))
    # For float years abuse the day variable
    fyear = year - year.astype('i')
    mask = fyear > 0
    if np.any(mask):
        year = year.astype('i')
        days_in_year = cal_to_jd(year[mask] + 1) - cal_to_jd(year[mask])
        day[mask] = days_in_year*fyear[mask]
        return _scalar_if_one(cal_to_jd(year) + day)
    return _scalar_if_one(cal_to_jd(year))


def yr_frac_mon_to_jd(year, mon, gregorian=True):
    """Convert a year and fractional month in the Julian or Gregorian
    calendars to the Julian Day Number (Meeus 7.1).

    Arguments:
      - `year` : (int)  year
      - `mon` : (int, float)  month

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (float)

    """
    year = np.atleast_1d(year)
    mon = np.atleast_1d(mon).astype(np.float64)
    day = np.atleast_1d(0.0).astype(np.float64)
    year, mon, day = map(np.array, np.broadcast_arrays(year,
                         mon, day))
    fmon = mon - mon.astype('i')
    mask = fmon > 0
    if np.any(mask):
        mon = mon.astype('i')
        next_mon = np.copy(mon) + 1
        next_year = np.copy(year)
        next_mon_mask = next_mon == 13
        next_year[next_mon_mask] = next_year[next_mon_mask] + 1
        next_mon[next_mon_mask] = 1
        days_in_mon = cal_to_jd(next_year[mask],
                                next_mon[mask]) - cal_to_jd(year[mask],
                                                            mon[mask])
        day[mask] = days_in_mon*fmon[mask]
        return _scalar_if_one(cal_to_jd(year, mon) + day)
    return _scalar_if_one(cal_to_jd(year, mon))


def cal_to_jd(year, mon=1, day=1, gregorian=True):
    """Convert a date in the Julian or Gregorian calendars to the Julian Day
    Number (Meeus 7.1).

    Arguments:
      - `year` : (int)  year

    Keywords:
      - `mon`       : (int, default=1) month
      - `day`       : (int, float, default=1) day, may be fractional day
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (int, float)

    """
    year = np.atleast_1d(year)
    mon = np.atleast_1d(mon)
    day = np.atleast_1d(day).astype(np.float64)

    fyear = year - year.astype('i')
    mask = fyear > 0
    if np.any(mask):
        raise ValueError('Year must be integer. Use frac_yr_to_jd instead.')
    fmon = mon - mon.astype('i')
    mask = fmon > 0
    if np.any(mask):
        raise ValueError('Month must be integer. Use yr_frac_mon_to_jd instead.')
    if np.any(mon > 12) or np.any(mon < 1):
        raise ValueError('Month must be from 1 to 12')
    if np.any(day > 31) or np.any(day < 1):
        raise ValueError('Day must be from 1 to 31')
    year, mon, day = map(np.array, np.broadcast_arrays(year, mon, day))

    for thirtydays in [9, 4, 6, 11]:
        daytestarr = mon == thirtydays
        if np.any(day[daytestarr] > 30):
            raise ValueError('Day must be from 1 to 30')

    leapyeartest = np.atleast_1d(is_leap_year(year, gregorian))

    if np.any(np.logical_and(day[leapyeartest] > 29, mon[leapyeartest] == 2)):
        raise ValueError('Day must be from 1 to 29')
    if np.any(np.logical_and(day[~leapyeartest] > 28,
                             mon[~leapyeartest] == 2)):
        raise ValueError('Day must be from 1 to 28')

    testarr = mon <= 2
    year[testarr] -= 1
    mon[testarr] += 12
    if gregorian:
        A = (year / 100).astype(np.int64)
        B = 2 - A + (A / 4).astype(np.int64)
    else:
        B = 0
    return _scalar_if_one(
        (365.25*(year + 4716)).astype(np.int64) +
        (30.6001*(mon + 1)).astype(np.int64) + day + B - 1524.5)


def jd_to_cal(julian_day, gregorian=True):
    """Convert a Julian day number to a date in the Julian or Gregorian
    calendars.

    Arguments:
      - `julian_day` : (int) Julian Day Number

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Return:
      - (year, month, day) : (tuple) day may be fractional

    """
    julian_day = np.atleast_1d(julian_day)
    F, Z = np.modf(julian_day + 0.5)
    if gregorian:
        alpha = ((Z - 1867216.25) / 36524.25).astype(np.int64)
        A = Z + 1 + alpha - (alpha / 4).astype(np.int64)
    else:
        A = Z
    B = A + 1524
    C = ((B - 122.1) / 365.25).astype(np.int64)
    D = (365.25 * C).astype(np.int64)
    E = ((B - D) / 30.6001).astype(np.int64)
    day = B - D - (30.6001 * E).astype(np.int64) + F
    mon = E - 13
    mon[E < 14] = E[E < 14] - 1
    year = C - 4715
    year[mon > 2] = C[mon > 2] - 4716
    return _scalar_if_one(year), _scalar_if_one(mon), _scalar_if_one(day)


def cal_to_jde(year, mon=1, day=1, hour=0, minute=0, sec=0.0, gregorian=True):
    """Convert a date in the Julian or Gregorian calendars to the Julian Day
    Ephemeris (Meeus 22.1).

    Arguments:
      - `year` : year

    Keywords:
      - `mon`       : (int, default=1) month
      - `day`       : (int, default=1) day, may be fractional day
      - `hour`      : (int, default=0) hour
      - `minute`    : (int, default=0) minute
      - `sec`       : (float, default=0.0) second
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - julian day ephemeris : (float)

    """
    jde = cal_to_jd(year, mon, day, gregorian)
    return _scalar_if_one(jde + hms_to_fday(hour, minute, sec))


def cal_to_day_of_year(year, mon, day, gregorian=True):
    """Convert a date in the Julian or Gregorian calendars to day of the year
    (Meeus 7.1).

    Arguments:
      - `year` : (int) year
      - `mon`  : (int) month
      - `day`  : (int) day

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Return:
      - day number : 1 = Jan 1...365 (or 366 for leap years) = Dec 31.

    """
    year = np.atleast_1d(year).astype(np.int64)
    mon = np.atleast_1d(mon).astype(np.int64)
    day = np.atleast_1d(day).astype(np.int64)
    year, mon, day = np.broadcast_arrays(year, mon, day)
    K = np.ones_like(year)
    K[:] = 2
    K[np.atleast_1d(is_leap_year(year, gregorian))] = 1
    return _scalar_if_one(
        (275 * mon / 9.0).astype(np.int64) -
        (K * ((mon + 9) / 12.0).astype(np.int64)) + day - 30)


def day_of_year_to_cal(year, N, gregorian=True):
    """Convert a day of year number to a month and day in the Julian or
    Gregorian calendars.

    Arguments:
      - `year`      : year
      - `N`         : day of year, 1..365 (or 366 for leap years)

    Keywords:
      - `gregorian` : If True, use Gregorian calendar, else use Julian calendar
        (default: True)

    Return:
      - (month, day) : (tuple)

    """
    year = np.atleast_1d(year)
    N = np.atleast_1d(N)
    year, N = np.broadcast_arrays(year, N)
    K = np.ones_like(N)
    K[:] = 2
    K[np.atleast_1d(is_leap_year(year, gregorian))] = 1
    mon = (9 * (K + N) / 275.0 + 0.98).astype(np.int64)
    mon[N < 32] = 1
    day = (N - (275 * mon / 9.0).astype(np.int64) +
           K * ((mon + 9) / 12.0).astype(np.int64) + 30).astype(np.int64)
    return _scalar_if_one(mon), _scalar_if_one(day)


def easter(year, gregorian=True):
    """Return the date of Western ecclesiastical Easter for a year in the
    Julian or Gregorian calendars.

    Arguments:
      - `year` : (int) year

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Return:
      - (month, day) : (tuple)

    """
    year = np.atleast_1d(year)
    if gregorian:
        a = year % 19
        b = year // 100
        c = year % 100
        d = b // 4
        e = b % 4
        f = (b + 8) // 25
        g = (b - f + 1) // 3
        h = (19 * a + b - d - g + 15) % 30
        i = c // 4
        k = c % 4
        l = (32 + 2 * e + 2 * i - h - k) % 7
        m = (a + 11 * h + 22 * l) // 451
        tmp = h + l - 7 * m + 114
    else:
        a = year % 4
        b = year % 7
        c = year % 19
        d = (19 * c + 15) % 30
        e = (2 * a + 4 * b - d + 34) % 7
        tmp = d + e + 114
    mon = tmp // 31
    day = (tmp % 31) + 1
    return _scalar_if_one(mon), _scalar_if_one(day)


def fday_to_hms(day):
    """Convert fractional day (0.0..1.0) to integral hours, minutes, seconds.

    Arguments:
      - day : a fractional day in the range 0.0..1.0

    Returns:
      - hour : (int, 0..23)
      - minute : (int, 0..59)
      - second : (int, 0..59)

    """
    # First get rid of the integer day
    fday, days = modf(day)
    seconds = fday * 86400.0
    minutes = int(seconds / 60.0)
    seconds = seconds - (minutes * 60.0)
    hours = int(minutes / 60.0)
    minutes = minutes - (hours * 60.0)
    return int(hours), int(minutes), int(seconds)


def hms_to_fday(hr, mn, sec):
    """Convert hours-minutes-seconds into a fractional day 0.0..1.0.

    Arguments:
      - `hr` : hours, 0..23
      - `mn` : minutes, 0..59
      - `sec` : seconds, 0..59

    Returns:
      - fractional day, 0.0..1.0

    """
    hr = np.atleast_1d(hr)
    mn = np.atleast_1d(mn)
    sec = np.atleast_1d(sec)
    hr, mn, sec = np.broadcast_arrays(hr, mn, sec)
    return ((hr / 24.0) + (mn / minutes_per_day) + (sec / seconds_per_day))


def is_dst(julian_day):
    """Is this instant within the Daylight Savings Time period.

    Uses the time zone database associated with Python and used in the 'time'
    module.

    Arguments:
      - `julian_day` : (int) Julian Day number representing an instant in
        Universal Time

    Returns:
      - (bool) True if Daylight Savings Time is in effect, False otherwise.

    """
    import time
    import datetime

    year, mon, day = jd_to_cal(julian_day)
    day = int(day)
    hr, minute, second = fday_to_hms(julian_day)
    second = int(second)

    stamp = datetime.datetime(year, mon, day, hr, minute, second)
    return time.localtime(time.mktime(stamp.timetuple())).tm_isdst == 1


def is_leap_year(year, gregorian=True):
    """Return True if this is a leap year in the Julian or Gregorian calendars

    Arguments:
      - `year` : (int) year

    Keywords:
      - `gregorian` : (bool, default=True) If True, use Gregorian calendar,
        else use Julian calendar

    Returns:
      - (bool) True is this is a leap year, else False.

    """
    year = np.atleast_1d(year).astype(np.int64)
    x = np.fmod(year, 4)
    if gregorian:
        x = np.fmod(year, 4)
        y = np.fmod(year, 100)
        z = np.fmod(year, 400)
        return _scalar_if_one(
            np.logical_and(np.logical_not(x),
                           np.logical_or(y, np.logical_not(z))))
    else:
        return _scalar_if_one(x == 0)


def jd_to_day_of_week(julian_day):
    """Return the day of week for a Julian Day Number.

    The Julian Day Number must be for 0h UT.

    Arguments:
      - `julian_day` : (int) Julian Day number

    Returns:
      - day of week : (int) 0 = Sunday...6 = Saturday.

    """
    julian_day = np.atleast_1d(julian_day)
    i = (julian_day + 1.5).astype(np.int64)
    return _scalar_if_one(i % 7)


def jd_to_jcent(julian_day):
    """Return the number of Julian centuries since J2000.0

    Arguments:
      - `julian_day` : (int) Julian Day number

    Return:
      - Julian centuries : (int)

    """
    julian_day = np.atleast_1d(julian_day)
    return _scalar_if_one((julian_day - 2451545.0) / 36525.0)


def lt_to_str(julian_day, zone="", level="second"):
    """Convert local time in Julian Days to a formatted string.

    The general format is:

        YYYY-MMM-DD HH:MM:SS ZZZ

    Truncate the time value to seconds, minutes, hours or days as
    indicated. If level = "day", don't print the time zone string.

    Pass an empty string ("", the default) for zone if you want to do
    your own zone formatting in the calling module.

    Arguments:
      - `julian_day` : (int) Julian Day number

    Keywords:
      - `zone`  : (str, default="") Time zone string
      - level : (str, default="second") {"day", "hour", "minute", "second"}

    Return:
      - formatted date/time string : (str)

    """
    year, mon, day = jd_to_cal(julian_day)
    fday, iday = modf(day)
    iday = int(iday)
    hour, minute, sec = fday_to_hms(fday)
    sec = int(sec)

    if level == "second":
        return "%d-%02d-%02d %02d:%02d:%02d %s" % (year, mon, iday,
                                                   hour, minute, sec,
                                                   zone)
    if level == "minute":
        return "%d-%02d-%02d %02d:%02d %s" % (year, mon, iday,
                                              hour, minute,
                                              zone)
    if level == "hour":
        return "%d-%02d-%02d %02d %s" % (year, mon, iday,
                                         hour,
                                         zone)
    if level == "day":
        return "%d-%02d-%02d" % (year, mon, iday)

    raise Error("unknown time level = " + level)


def sidereal_time_greenwich(julian_day):
    """Return the mean sidereal time at Greenwich.

    The Julian Day number must represent Universal Time.

    Arguments:
      - `julian_day` : (int) Julian Day number

    Returns:
      - sidereal time in radians : (float) 2pi radians = 24 hours

    """
    T = jd_to_jcent(julian_day)
    T2 = T * T
    T3 = T2 * T
    theta0 = 280.46061837 + \
        360.98564736629*(julian_day - 2451545.0) + \
        0.000387933*T2 - \
        T3/38710000
    result = d_to_r(theta0)
    return modpi2(result)


def ut_to_lt(julian_day):
    """Convert universal time in Julian Days to a local time.

    Include Daylight Savings Time offset, if any.

    Arguments:
      - `julian_day` : (int) Julian Day number, universal time

    Return:
      - Julian Day number : (str) local time
        zone string of the zone used for the conversion

    """
    if is_dst(julian_day):
        zone = astronomia.globals.daylight_timezone_name
        offset = astronomia.globals.daylight_timezone_offset
    else:
        zone = astronomia.globals.standard_timezone_name
        offset = astronomia.globals.standard_timezone_offset

    julian_day = julian_day - offset
    return julian_day, zone



#! /usr/bin/env python

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

CGI interface to the Easter routine.

Generate dates of Easter for a range of years, both Gregorian
and Julian calendars.

One application handles both the input form and the displayed
results.

"""

import cgi
import sys
import time

#
# Rather than setup PYTHONPATH in a shell script, we add it
# through sys. That way we need only one file for the application.
#
# This path is specific to SourceForge.
#
sys.path.append('/home/groups/a/as/astronomia/lib/python')

from astronomia.calendar import easter

form = cgi.FieldStorage()

#
# Collect the input parameters
#
def display_form():
    fields = time.localtime(time.time())
    year = fields[0]

    print('Content-type: text/html')
    print()
    print('<HEAD>')
    print('<TITLE>Astronomia Easter Parameters</TITLE>')
    print('</HEAD>')

    print('<BODY>')
    print('<H1>Astronomia Easter Parameters</H1>')

    print('<FORM action="http://astronomia.sourceforge.net/cgi-bin/easter-cgi.py" method="post">')

    print('<TABLE>')
    print('<TR>')
    print('<TD><LABEL for="starting_year">Starting year</LABEL></TD>')
    print('<TD><INPUT type="text" size=4 name="starting_year" id="starting_year" value="%d"></TD>' % (year - 5))
    print('</TR>')
    print('<TR>')
    print('<TD><LABEL for="ending_year">Ending year</LABEL></TD>')
    print('<TD><INPUT type="text" size=4 name="ending_year" id="ending_year" value="%d"></TD>' % (year + 5))
    print('</TR>')
    print('</TABLE>')

    print('<INPUT type="checkbox" checked name="gregorian_calendar" value="1"> Gregorian calendar<BR>')
    print('<INPUT type="checkbox" checked name="julian_calendar" value="1"> Julian calendar<BR>')

    print('<INPUT type="submit" value="Run"> <INPUT type="reset">')

    print('<INPUT type="hidden" name="results" value="1">')

    print('</FORM>')
    print('</BODY>')

#
# Display an error message
#
def display_error(msg):
    print('<HEAD>')
    print('<TITLE>Astronomia Easter Error</TITLE>')
    print('</HEAD>')
    print('<BODY>')
    print('<H1>', msg, '</H1>')
    print('</BODY>')

#
# Display the results
#
def display_results():
    print('Content-type: text/html')
    print()
    print('<HEAD>')
    print('<TITLE>Astronomia Easter Results</TITLE>')
    print('</HEAD>')

    #
    # To do: We should validate the integer values
    #
    # If no starting_year, start with 5 years ago.
    #
    if 'starting_year' in form:
        starting_year = int(form['starting_year'].value)
    else:
        fields = time.localtime(time.time())
        year = fields[0]
        starting_year = year - 5

    #
    # If no ending_year, end 10 years after the starting_year.
    #
    if 'ending_year' in form:
        ending_year = int(form['ending_year'].value)
    else:
        ending_year = starting_year + 10

    #
    # Make sure end_year >= starting_year.
    #
    if (ending_year < starting_year):
        display_error('Ending year %d is earlier than starting year %d' % (ending_year, starting_year))
        return

    #
    # We can't go farther back than the first Julian Day number.
    #
    if (starting_year < -4712):
        display_error('Starting year %d is earlier than 4713BC' % starting_year)
        return

    julian_calendar = 'julian_calendar' in form
    gregorian_calendar = 'gregorian_calendar' in form

    #
    # Must have at least one calendar type.
    #
    if not julian_calendar and not gregorian_calendar:
        display_error('No calendar type selected')
        return

    #
    # Finally, we display the results.
    #
    print('<BODY>')
    print('<H1>Astronomia Easter Results</H1>')

    print('<TABLE BORDER>')

    print('<TR>')
    if gregorian_calendar:
        print('<TH>Gregorian</TH>')
    if julian_calendar:
        print('<TH>Julian</TH>')
    print('</TR>')

    for year in range(starting_year, ending_year + 1):
        print('<TR>')
        if gregorian_calendar:
            month, day = easter(year, True)
            print('<TD><TT>%02d-%02d-%d</TT></TD>' % (day, month, year))
        if julian_calendar:
            month, day = easter(year, False)
            print('<TD><TT>%02d-%02d-%d</TT></TD>' % (day, month, year))
        print('</TR>')
    print('</TABLE>')

    # handy diagnostic routines
    """
    print "<H1>print_environ</H1>"
    cgi.print_environ()
    print "<H1>print_directory</H1>"
    cgi.print_directory()
    print "<H1>print_environ_usage</H1>"
    cgi.print_environ_usage()
    print "<H1>FieldStorage</H1>"
    #form = cgi.FieldStorage()
    cgi.print_form(form)
    """

    print('</BODY>')

#
# Main routine. There is a hidden field "results" on the parameter form. If it
# is not present, we display the form. If it is, we display the results.
#
if 'results' in form:
    display_results()
else:
    display_form()






#! /usr/bin/env python

from __future__ import print_function

"""
    Astrolabe copyright 2000, 2001 William McClain
    Astrolabe forked to Astronomia 2013
    Astronomia copyright 2013

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

CGI interface to the Equinox/Solstice routines.

Will display a table of times of equinoxes and solstices for a
range of years.

One application handles both the input form and the displayed
results.

"""

import cgi
import sys
import time

#
# Rather than setup PYTHONPATH in a shell script, we add it
# through sys. That way we need only one file for the application.
#
#
# This path is specific to SourceForge.
#
sys.path.append('/home/groups/a/as/astronomia/lib/python')

from astronomia.constants import days_per_second
from astronomia.calendar import ut_to_lt, lt_to_str
from astronomia.dynamical import dt_to_ut
from astronomia.equinox import equinox_approx, equinox
from astronomia.sun import Sun
from astronomia.util import load_params
import astronomia.globals

standard_zones = (
    'GMT-12',
    'GMT-11',
    'GMT-10',
    'GMT-9',
    'PST',
    'MST',
    'CST',
    'EST',
    'GMT-4',
    'GMT-3',
    'GMT-2',
    'GMT-1',
    'GMT',
    'GMT+1',
    'GMT+2',
    'GMT+3',
    'GMT+4',
    'GMT+5',
    'GMT+6',
    'GMT+7',
    'GMT+8',
    'GMT+9',
    'GMT+10',
    'GMT+11',
    'GMT+12')

standard_zone_hours = {
    'GMT-12': -12,
    'GMT-11': -11,
    'GMT-10': -10,
    'GMT-9': -9,
    'PST': -8,
    'MST': -7,
    'CST': -6,
    'EST': -5,
    'GMT-4': -4,
    'GMT-3': -3,
    'GMT-2': -2,
    'GMT-1': -1,
    'GMT': 0,
    'GMT+1': 1,
    'GMT+2': 2,
    'GMT+3': 3,
    'GMT+4': 4,
    'GMT+5': 5,
    'GMT+6': 6,
    'GMT+7': 7,
    'GMT+8': 8,
    'GMT+9': 9,
    'GMT+10': 10,
    'GMT+11': 11,
    'GMT+12': 12}

daylight_zones = (
    'GMT-12',
    'GMT-11',
    'GMT-10',
    'GMT-9',
    'GMT-8',
    'PDT',
    'MDT',
    'CDT',
    'EDT',
    'GMT-3',
    'GMT-2',
    'GMT-1',
    'GMT',
    'GMT+1',
    'GMT+2',
    'GMT+3',
    'GMT+4',
    'GMT+5',
    'GMT+6',
    'GMT+7',
    'GMT+8',
    'GMT+9',
    'GMT+10',
    'GMT+11',
    'GMT+12')

daylight_zone_hours = {
    'GMT-12': -12,
    'GMT-11': -11,
    'GMT-10': -10,
    'GMT-9': -9,
    'GMT-8': -8,
    'PDT': -7,
    'MDT': -6,
    'CDT': -5,
    'EDT': -4,
    'GMT-3': -3,
    'GMT-2': -2,
    'GMT-1': -1,
    'GMT': 0,
    'GMT+1': 1,
    'GMT+2': 2,
    'GMT+3': 3,
    'GMT+4': 4,
    'GMT+5': 5,
    'GMT+6': 6,
    'GMT+7': 7,
    'GMT+8': 8,
    'GMT+9': 9,
    'GMT+10': 10,
    'GMT+11': 11,
    'GMT+12': 12}

default_standard_zone = 'CST'
default_daylight_zone = 'CDT'

form = cgi.FieldStorage()

#
# Collect the input parameters
#
def display_form():
    fields = time.localtime(time.time())
    year = fields[0]
    
    print('Content-type: text/html') 
    print()                               
    print('<HEAD>')
    print('<TITLE>Astronomia Solstice Parameters</TITLE>')
    print('</HEAD>')
    
    print('<BODY>')
    print('<H1>Astronomia Solstice Parameters</H1>')
    
    print('<FORM action="http://astronomia.sourceforge.net/cgi-bin/solstice-cgi.py" method="post">')
    
    print('<TABLE>')
    print('<TR>')
    print('<TD><LABEL for="starting_year">Starting year</LABEL></TD>')
    print('<TD><INPUT type="text" size=4 name="starting_year" id="starting_year" value="%d"></TD>' % (year - 5))
    print('</TR>')
    print('<TR>')
    print('<TD><LABEL for="ending_year">Ending year</LABEL></TD>')
    print('<TD><INPUT type="text" size=4 name="ending_year" id="ending_year" value="%d"></TD>' % (year + 5))
    print('</TR>')
    print('<TR>')
    print('<TD><LABEL for="standard_zone">Standard Time Zone</LABEL></TD>')
    print('<TD><SELECT name = "standard_zone">')
    for zone in standard_zones:
        if zone == default_standard_zone:
            print('<OPTION SELECTED>' + zone + '</OPTION>')
        else:
            print('<OPTION>' + zone + '</OPTION>')
    print('</SELECT></TD>')
    print('</TR>')
    print('<TR>')
    print('<TD><LABEL for="daylight_zone">Daylight Time Zone</LABEL></TD>')
    print('<TD><SELECT name = "daylight_zone">')
    for zone in daylight_zones:
        if zone == default_daylight_zone:
            print('<OPTION SELECTED>' + zone + '</OPTION>')
        else:
            print('<OPTION>' + zone + '</OPTION>')
    print('</SELECT></TD>')
    print('</TR>')
    print('</TABLE>')

    print('<INPUT type="submit" value="Run"> <INPUT type="reset">')
 
    print('<INPUT type="hidden" name="results" value="1">')
 
    print('</FORM>')
    print('</BODY>')

#
# Display an error message
#
def display_error(msg):
    print('<HEAD>')
    print('<TITLE>Astronomia Solstice Error</TITLE>')
    print('</HEAD>')
    print('<BODY>')
    print('<H1>', msg, '</H1>')
    print('</BODY>')

#
# Display the results
#
def display_results():
    #
    # Assign the global parameters directly so we don't have to read a param file.
    #
    astronomia.globals.standard_timezone_name = form['standard_zone'].value
    astronomia.globals.standard_timezone_offset = -standard_zone_hours[astronomia.globals.standard_timezone_name] / 24.0
    astronomia.globals.daylight_timezone_name = form['daylight_zone'].value
    astronomia.globals.daylight_timezone_offset = -daylight_zone_hours[astronomia.globals.daylight_timezone_name] / 24.0
    astronomia.globals.vsop87d_text_path = '/home/groups/a/as/astronomia/data/vsop87d.txt'
    astronomia.globals.vsop87d_binary_path = '/home/groups/a/as/astronomia/data/vsop87d.dat'

    print('Content-type: text/html') 
    print()                               
    print('<HEAD>')
    print('<TITLE>Astronomia Solstice Results</TITLE>')
    print('</HEAD>')

    #
    # To do: We should validate the integer values
    #
    # If no starting_year, start with 5 years ago.
    #
    if 'starting_year' in form:
        starting_year = int(form['starting_year'].value)
    else:
        fields = time.localtime(time.time())
        year = fields[0]
        starting_year = year - 5

    #
    # If no ending_year, end 10 years after the starting_year.
    #
    if 'ending_year' in form:
        ending_year = int(form['ending_year'].value)
    else:
        ending_year = starting_year + 10

    #
    # Make sure end_year >= starting_year.
    #
    if (ending_year < starting_year):
        display_error('Ending year %d is earlier than starting year %d' % (ending_year, starting_year))
        return

    #
    # We can't go farther back than the first Julian Day number.
    #
    if (starting_year < -4712):
        display_error('Starting year %d is earlier than 4713BC' % starting_year)
        return

    #
    # Finally, we display the results.
    #
    print('<BODY>')
    print('<H1>Astronomia Solstice Results</H1>')

    print('<TABLE BORDER>')

    print('<TR>')
    print('<TH>Vernal Equinox</TH>')
    print('<TH>Summer Solstice</TH>')
    print('<TH>Autumnal Equinox</TH>')
    print('<TH>Winter Solstice</TH>')
    print('</TR>')

    for year in range(starting_year, ending_year + 1):
        print('<TR>')
        for season in astronomia.globals.season_names:
            approx_jd = equinox_approx(year, season)
            jd = equinox(approx_jd, season, days_per_second)
            ut = dt_to_ut(jd)
            lt, zone = ut_to_lt(ut)
            print('<TD>' + lt_to_str(lt, zone) + '</TD>')
        print('</TR>')
    print('</TABLE>')

    # handy diagnostic routines
    """
    print "<H1>print_environ</H1>"
    cgi.print_environ() 
    print "<H1>print_directory</H1>"
    cgi.print_directory() 
    print "<H1>print_environ_usage</H1>"
    cgi.print_environ_usage()
    print "<H1>FieldStorage</H1>"
    #form = cgi.FieldStorage()
    cgi.print_form(form)
    """
    
    print('</BODY>')

#
# Main routine. There is a hidden field "results" on the parameter form. If it
# is not present, we display the form. If it is, we display the results.
#
if 'results' in form:
    display_results()
else:
    display_form()






"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    """

"""Useful constants terms.

Don't change these unless you are moving to a new universe.

"""
from astronomia.util import d_to_r

#
# Constant terms.
#
kL1 = (d_to_r(218.3164477),
       d_to_r(481267.88123421),
       d_to_r(-0.0015786),
       d_to_r(1.0/538841),
       d_to_r(-1.0/65194000))
kD = (d_to_r(297.8501921),
      d_to_r(445267.1114034),
      d_to_r(-0.0018819),
      d_to_r(1.0/545868),
      d_to_r(-1.0/113065000))
kM = (d_to_r(357.5291092),
      d_to_r(35999.0502909),
      d_to_r(-0.0001536),
      d_to_r(1.0/24490000))
kM1 = (d_to_r(134.9633964),
       d_to_r(477198.8675055),
       d_to_r(0.0087414),
       d_to_r(1.0/69699),
       d_to_r(-1.0/14712000))
kF = (d_to_r(93.2720950),
      d_to_r(483202.0175233),
      d_to_r(-0.0036539),
      d_to_r(-1.0/3526000),
      d_to_r(1.0/863310000))
ko = (d_to_r(125.0445479),
      d_to_r(-1934.1362891),
      d_to_r(0.0020754),
      d_to_r(1.0/467441.0),
      d_to_r(1.0/60616000.0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py upload_docs')
    sys.exit()

readme = open('README.rst').read()

setup(name='astronomia',
      version=open("VERSION").readline().strip(),
      description="Library for calculation of ephemeris and other astronomical calculations",
      long_description=readme + '\n\n',
      classifiers=['Environment :: Console',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.6',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.3',
                   'Topic :: Scientific/Engineering :: Astronomy',
                   ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='ephemeris astronomy',
      author='Tim Cera, P.E.',
      author_email='tim@cerazone.net',
      url='http://timcera.bitbucket.org/',
      license='GPL2',
      packages=['astronomia'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          ],
      scripts=['apps/solstice.py', 'apps/check_perihelion.py',
          'apps/cronus.py', 'apps/easter-cgi.py', 'apps/solstice-cgi.py'],
      data_files=[(os.path.join('share', 'astronomia'), ['apps/astronomia_params.txt'])],
      entry_points="""
      # -*- Entry points: -*-
      """,
      test_suite='tests',
      )



"""
    Copyright 2000, 2001 Astrolabe by William McClain

    Forked in 2013 to Astronomia

    Copyright 2013 Astronomia by Tim Cera

    This file is part of Astronomia.

    Astronomia is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Astronomia is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Astronomia; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    """

"""Lunar position model ELP2000-82 of Chapront.

The result values are for the equinox of date and have been adjusted
for light-time.

This is the simplified version of Jean Meeus, _Astronomical Algorithms_,
second edition, 1998, Willmann-Bell, Inc.

"""
import numpy as np

from astronomia.calendar import jd_to_jcent
from astronomia.util import polynomial, d_to_r, modpi2
from astronomia.commonterms import kL1, kD, kM, kM1, kF, ko


class Error(Exception):
    """Local exception class"""
    pass

# [Meeus-1998: table 47.A]
#
#    D, M, M1, F, l, r

_tblLR = (
    (0,  0,  1,  0, 6288774, -20905355),
    (2,  0, -1,  0, 1274027,  -3699111),
    (2,  0,  0,  0,  658314,  -2955968),
    (0,  0,  2,  0,  213618,   -569925),
    (0,  1,  0,  0, -185116,     48888),
    (0,  0,  0,  2, -114332,     -3149),
    (2,  0, -2,  0,   58793,    246158),
    (2, -1, -1,  0,   57066,   -152138),
    (2,  0,  1,  0,   53322,   -170733),
    (2, -1,  0,  0,   45758,   -204586),
    (0,  1, -1,  0,  -40923,   -129620),
    (1,  0,  0,  0,  -34720,    108743),
    (0,  1,  1,  0,  -30383,    104755),
    (2,  0,  0, -2,   15327,     10321),
    (0,  0,  1,  2,  -12528,         0),
    (0,  0,  1, -2,   10980,     79661),
    (4,  0, -1,  0,   10675,    -34782),
    (0,  0,  3,  0,   10034,    -23210),
    (4,  0, -2,  0,    8548,    -21636),
    (2,  1, -1,  0,   -7888,     24208),
    (2,  1,  0,  0,   -6766,     30824),
    (1,  0, -1,  0,   -5163,     -8379),
    (1,  1,  0,  0,    4987,    -16675),
    (2, -1,  1,  0,    4036,    -12831),
    (2,  0,  2,  0,    3994,    -10445),
    (4,  0,  0,  0,    3861,    -11650),
    (2,  0, -3,  0,    3665,     14403),
    (0,  1, -2,  0,   -2689,     -7003),
    (2,  0, -1,  2,   -2602,         0),
    (2, -1, -2,  0,    2390,     10056),
    (1,  0,  1,  0,   -2348,      6322),
    (2, -2,  0,  0,    2236,     -9884),
    (0,  1,  2,  0,   -2120,      5751),
    (0,  2,  0,  0,   -2069,         0),
    (2, -2, -1,  0,    2048,     -4950),
    (2,  0,  1, -2,   -1773,      4130),
    (2,  0,  0,  2,   -1595,         0),
    (4, -1, -1,  0,    1215,     -3958),
    (0,  0,  2,  2,   -1110,         0),
    (3,  0, -1,  0,    -892,      3258),
    (2,  1,  1,  0,    -810,      2616),
    (4, -1, -2,  0,     759,     -1897),
    (0,  2, -1,  0,    -713,     -2117),
    (2,  2, -1,  0,    -700,      2354),
    (2,  1, -2,  0,     691,         0),
    (2, -1,  0, -2,     596,         0),
    (4,  0,  1,  0,     549,     -1423),
    (0,  0,  4,  0,     537,     -1117),
    (4, -1,  0,  0,     520,     -1571),
    (1,  0, -2,  0,    -487,     -1739),
    (2,  1,  0, -2,    -399,         0),
    (0,  0,  2, -2,    -381,     -4421),
    (1,  1,  1,  0,     351,         0),
    (3,  0, -2,  0,    -340,         0),
    (4,  0, -3,  0,     330,         0),
    (2, -1,  2,  0,     327,         0),
    (0,  2,  1,  0,    -323,      1165),
    (1,  1, -1,  0,     299,         0),
    (2,  0,  3,  0,     294,         0),
    (2,  0, -1, -2,       0,      8752))

# [Meeus-1998: table 47.B]
#
#    D, M, M1, F, b

_tblB = (
    (0,  0,  0,  1, 5128122),
    (0,  0,  1,  1,  280602),
    (0,  0,  1, -1,  277693),
    (2,  0,  0, -1,  173237),
    (2,  0, -1,  1,   55413),
    (2,  0, -1, -1,   46271),
    (2,  0,  0,  1,   32573),
    (0,  0,  2,  1,   17198),
    (2,  0,  1, -1,    9266),
    (0,  0,  2, -1,    8822),
    (2, -1,  0, -1,    8216),
    (2,  0, -2, -1,    4324),
    (2,  0,  1,  1,    4200),
    (2,  1,  0, -1,   -3359),
    (2, -1, -1,  1,    2463),
    (2, -1,  0,  1,    2211),
    (2, -1, -1, -1,    2065),
    (0,  1, -1, -1,   -1870),
    (4,  0, -1, -1,    1828),
    (0,  1,  0,  1,   -1794),
    (0,  0,  0,  3,   -1749),
    (0,  1, -1,  1,   -1565),
    (1,  0,  0,  1,   -1491),
    (0,  1,  1,  1,   -1475),
    (0,  1,  1, -1,   -1410),
    (0,  1,  0, -1,   -1344),
    (1,  0,  0, -1,   -1335),
    (0,  0,  3,  1,    1107),
    (4,  0,  0, -1,    1021),
    (4,  0, -1,  1,     833),
    (0,  0,  1, -3,     777),
    (4,  0, -2,  1,     671),
    (2,  0,  0, -3,     607),
    (2,  0,  2, -1,     596),
    (2, -1,  1, -1,     491),
    (2,  0, -2,  1,    -451),
    (0,  0,  3, -1,     439),
    (2,  0,  2,  1,     422),
    (2,  0, -3, -1,     421),
    (2,  1, -1,  1,    -366),
    (2,  1,  0,  1,    -351),
    (4,  0,  0,  1,     331),
    (2, -1,  1,  1,     315),
    (2, -2,  0, -1,     302),
    (0,  0,  1,  3,    -283),
    (2,  1,  1, -1,    -229),
    (1,  1,  0, -1,     223),
    (1,  1,  0,  1,     223),
    (0,  1, -2, -1,    -220),
    (2,  1, -1, -1,    -220),
    (1,  0,  1,  1,    -185),
    (2, -1, -2, -1,     181),
    (0,  1,  2,  1,    -177),
    (4,  0, -2, -1,     176),
    (4, -1, -1, -1,     166),
    (1,  0,  1, -1,    -164),
    (4,  0,  1, -1,     132),
    (1,  0, -1, -1,    -119),
    (4, -1,  0, -1,     115),
    (2, -2,  0,  1,     107))

_kA1 = (d_to_r(119.75), d_to_r(131.849))
_kA2 = (d_to_r(53.09), d_to_r(479264.290))
_kA3 = (d_to_r(313.45), d_to_r(481266.484))


def _constants(T):
    """Calculate values required by several other functions"""
    L1 = modpi2(polynomial(kL1, T))
    D = modpi2(polynomial(kD, T))
    M = modpi2(polynomial(kM, T))
    M1 = modpi2(polynomial(kM1, T))
    F = modpi2(polynomial(kF, T))

    A1 = modpi2(polynomial(_kA1, T))
    A2 = modpi2(polynomial(_kA2, T))
    A3 = modpi2(polynomial(_kA3, T))

    E = polynomial([1.0, -0.002516, -0.0000074], T)
    E2 = E*E

    return L1, D, M, M1, F, A1, A2, A3, E, E2


class Lunar:
    """ELP2000 lunar position calculations"""

    def mean_longitude_ascending_node(self, jd):
        """Return mean longitude of ascending node

        Another equation from:
           This routine is part of the International Astronomical Union's
           SOFA (Standards of Fundamental Astronomy) software collection.
           Fundamental (Delaunay) arguments from Simon et al. (1994)

        *  Arcseconds to radians
           DOUBLE PRECISION DAS2R
           PARAMETER ( DAS2R = 4.848136811095359935899141D-6 )

        *  Milliarcseconds to radians
           DOUBLE PRECISION DMAS2R
           PARAMETER ( DMAS2R = DAS2R / 1D3 )

        *  Arc seconds in a full circle
           DOUBLE PRECISION TURNAS
           PARAMETER ( TURNAS = 1296000D0 )

        *  Mean longitude of the ascending node of the Moon.
           OM  = MOD ( 450160.398036D0  -6962890.5431D0*T, TURNAS ) * DAS2R

        Keeping above for documentation, but...
        Current implemention in astronomia is from:

           PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

        Arguments:
          - `jd` : julian Day

        Returns:
          - mean longitude of ascending node
        """
        T = jd_to_jcent(jd)
        return modpi2(polynomial(ko, T))

    def mean_longitude_perigee(self, jd):
        """Return mean longitude of lunar perigee

        Arguments:
          - `jd` : julian Day

        Returns:
          - mean longitude of perigee

        """
        T = jd_to_jcent(jd)
        X = polynomial(
            (d_to_r(83.3532465),
             d_to_r(4069.0137287),
             d_to_r(-0.0103200),
             d_to_r(-1./80053),
             d_to_r(1./18999000)
             ), T)
        return modpi2(X)

    def mean_longitude(self, jd):
        """Return geocentric mean longitude.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - mean longitude in radians

        """
        T = jd_to_jcent(jd)
        return modpi2(polynomial(kL1, T))

    def mean_elongation(self, jd):
        """Return geocentric mean elongation.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - mean elongation in radians

        """
        T = jd_to_jcent(jd)
        return modpi2(polynomial(kD, T))

    def mean_anomaly(self, jd):
        """Return geocentric mean anomaly.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - mean anomaly in radians

        """
        T = jd_to_jcent(jd)
        return modpi2(polynomial(kM1, T))

    def argument_of_latitude(self, jd):
        """Return geocentric mean longitude.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - argument of latitude in radians

        """
        T = jd_to_jcent(jd)
        return modpi2(polynomial(kF, T))

    def dimension3(self, jd):
        """Return geocentric ecliptic longitude, latitude and radius.

        When we need all three dimensions it is more efficient to combine the
        calculations in one routine.

        Arguments:
          - `jd` : Julian Day in dynamical time

        Returns:
          - longitude in radians
          - latitude in radians
          - radius in km, Earth's center to Moon's center

        """
        return self._longitude(jd), self._latitude(jd), self._radius(jd)

    def dimension(self, jd, dim):
        """Return one of geocentric ecliptic longitude, latitude and radius.

        Arguments:
          - `jd` : Julian Day in dynamical time
          - `dim` : "L" (longitude") or "B" (latitude) or "R" (radius)

        Returns:
          - longitude in radians or latitude in radians or radius in km,
            Earth's center to Moon's center, depending on value of `dim`.

        """
        if dim == "L":
            return self._longitude(jd)
        if dim == "B":
            return self._latitude(jd)
        if dim == "R":
            return self._radius(jd)
        raise Error("unknown dimension = " + dim)

    def _longitude(self, jd):
        """Return the geocentric ecliptic longitude in radians.
        """
        from astronomia.nutation import nutation_in_longitude

        T = jd_to_jcent(jd)
        L1, D, M, M1, F, A1, A2, A3, E, E2 = _constants(T)
        lsum = 0.0
        for tD, tM, tM1, tF, tl, tr in _tblLR:
            arg = tD * D + tM * M + tM1 * M1 + tF * F
            if abs(tM) == 1:
                tl *= E
            elif abs(tM) == 2:
                tl *= E2
            lsum += tl * np.sin(arg)

        lsum += 3958*np.sin(A1) + 1962*np.sin(L1 - F) + 318*np.sin(A2)

        nutinlong = nutation_in_longitude(jd)
        longitude = L1 + d_to_r(lsum / 1000000) + nutinlong
        return longitude

    def _latitude(self, jd):
        """Return the geocentric ecliptic latitude in radians.
        """
        T = jd_to_jcent(jd)
        L1, D, M, M1, F, A1, A2, A3, E, E2 = _constants(T)

        bsum = 0.0
        for tD, tM, tM1, tF, tb in _tblB:
            arg = tD * D + tM * M + tM1 * M1 + tF * F
            if abs(tM) == 1:
                tb *= E
            elif abs(tM) == 2:
                tb *= E2
            bsum += tb * np.sin(arg)

        bsum += -2235 * np.sin(L1) +    \
            382 * np.sin(A3) +      \
            175 * np.sin(A1 - F) +  \
            175 * np.sin(A1 + F) +  \
            127 * np.sin(L1 - M1) - \
            115 * np.sin(L1 + M1)

        latitude = d_to_r(bsum / 1000000)
        return latitude

    def _radius(self, jd):
        """Return the geocentric radius in km.
        """
        T = jd_to_jcent(jd)
        L1, D, M, M1, F, A1, A2, A3, E, E2 = _constants(T)

        rsum = 0.0
        for tD, tM, tM1, tF, tl, tr in _tblLR:
            arg = tD*D + tM*M + tM1*M1 + tF*F
            if abs(tM) == 1:
                tr *= E
            elif abs(tM) == 2:
                tr *= E2
            rsum += tr * np.cos(arg)

        dist = 385000.56 + rsum / 1000
        return dist
